'use strict';

var CODE_SUCCESS = 200;
var CODE_FAIL = 300;
var requests = {};

function instanceOf(src, name) {
  try {
    return src instanceof window[name];
  } catch (e) {
    return false;
  }
}

function parseRequestData(params) {
  if (
  // ArrayBufferView
  instanceOf(params, 'Int8Array') || instanceOf(params, 'Uint8Array') || instanceOf(params, 'Uint8ClampedArray') || instanceOf(params, 'Int16Array') || instanceOf(params, 'Uint16Array') || instanceOf(params, 'Int32Array') || instanceOf(params, 'Uint32Array') || instanceOf(params, 'Float32Array') || instanceOf(params, 'Float64Array') || instanceOf(params, 'DataView') ||
  // Blob
  instanceOf(params, 'Blob') ||
  // Document
  instanceOf(params, 'HTMLDocument') || instanceOf(params, 'XMLDocument') ||
  // DOMString
  instanceOf(params, 'String') || typeof params === 'string' ||
  // FormData
  instanceOf(params, 'FormData')) {
    return params;
  }

  return JSON.stringify(params);
}

function parseResponseData(data, defaults) {
  try {
    return JSON.parse(data);
  } catch (e) {
    return defaults;
  }
}

function makeGetUrl(url, data) {
  return url + (url.indexOf('?') !== -1 ? '&' : '?') + Object.keys(data).map(function (key) {
    return encodeURIComponent(key) + '=' + encodeURIComponent(data[key]);
  }).join('&');
}

function request(method, url, data, headers, withCredentials) {
  var xhr = new XMLHttpRequest();
  xhr.withCredentials = withCredentials;

  var promise = new Promise(function (resolve, reject) {
    xhr.open(method, url, true);

    Object.keys(headers).forEach(function (name) {
      return xhr.setRequestHeader(name, headers[name]);
    });

    xhr.addEventListener('timeout', reject);
    xhr.addEventListener('abort', reject);
    xhr.addEventListener('error', reject);
    xhr.addEventListener('load', function () {
      requests[promise.url].splice(requests[promise.url].indexOf(promise), 1);

      if (CODE_SUCCESS <= xhr.status && xhr.status < CODE_FAIL) {
        resolve({
          status: xhr.status,
          data: parseResponseData(xhr.response, xhr.response)
        });
      } else {
        reject({
          status: xhr.status,
          data: parseResponseData(xhr.response, null)
        });
      }
    });
  });

  promise.abort = xhr.abort.bind(xhr);

  xhr.send(data);

  return promise;
}

function add(url, request) {
  request.url = url;
  requests[url] = requests[url] || [];
  requests[url].push(request);
  return request;
}

var ajax = {
  headers: {},

  withCredentials: true,

  get: function get(url, data, headers) {
    return add(url, request('GET', data ? makeGetUrl(url, data) : url, null, Object.assign({}, this.headers, headers), this.withCredentials));
  },
  post: function post(url, data, headers) {
    return add(url, request('POST', url, parseRequestData(data), Object.assign({}, this.headers, headers), this.withCredentials));
  },
  put: function put(url, data, headers) {
    return add(url, request('PUT', url, parseRequestData(data), Object.assign({}, this.headers, headers), this.withCredentials));
  },
  patch: function patch(url, data, headers) {
    return add(url, request('PATCH', url, parseRequestData(data), Object.assign({}, this.headers, headers), this.withCredentials));
  },
  delete: function _delete(url, data, headers) {
    return add(url, request('DELETE', url, parseRequestData(data), Object.assign({}, this.headers, headers), this.withCredentials));
  },
  stop: function stop(url) {
    (requests[url] || []).forEach(function (request) {
      return request.abort();
    });
    requests[url] = [];
  }
};

module.exports = {
  install: function install(Vue) {
    Vue.http = ajax;
    Object.defineProperty(Vue.prototype, '$http', { get: function get() {
        return ajax;
      } });
  }
};
